SystemOrganization addCategory: #'XBase-Base'!

Object subclass: #XBaseFile
	instanceVariableNames: 'filterArray lastUpdate fileName fileStream headerSize recordSize nRecords currentRecordIndex deleted modified eof bof recordBytes readBlocks writeBlocks fields nFields fieldFormats fieldIndices'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XBase-Base'!

!XBaseFile commentStamp: 'jmb 11/26/2004 12:36' prior: 0!
A Smalltalk class giving full access to Xbase (dBASE) files (dBASE III/IV). Hans Baveco, November 3, 2004.
A rewrite of a Squeak goodie (XBase, version 1.00, published Feb 13, 1998 by andavino@tin.it). Ported to VisualWorks, back-ported to Squeak!!


Open a dbf file with:
	| dbf | 
	dbf := XBaseFile on: 'd:\DBF\allotments.dbf'. 
"or"
	dbf := XBaseFile fromUser.
	dbf inspect.
"Close it with:"
	dbf close
	
Retrieve the contents of a dbf with:
	XBaseFile recordsFromFile: 'd:\DBF\allotments.dbf'. 
or
	XBaseFile recordsFromUser.
The file does not need to be explicitly closed now.
If you want to have the field(labels) as well:
	XBaseFile recordsAndFieldsFromFile: 'd:\DBF\allotments.dbf'. 
or
	XBaseFile recordsAndFieldsFromUser.


Create a new dbf file with the on:fields: method, e.g.:
	| dbf rng |
	dbf := XBaseFile on: 'testEdit' fields: 
		{#( 'abc' $N 5 0 nil). #( 'et_class' $C 12 0 nil). #( 'perimeter' $F 20 3 nil). #( 'logical' $L 1 0 nil)}.
"4 fields, but still empty. Add records:"
	rng := Random new. rng seed: 2345678901.
	1 to: 10
		do: 
			[:ix | 
			dbf append.
			dbf atFieldName: 'ABC' put: ix.
			dbf atFieldName: 'ET_CLASS' put: 'MyClass', ix printString.
			dbf atFieldName: 'PERIMETER' put: ix  * 123456.12345.
			dbf atFieldName: 'LOGICAL' put: (rng next > 0.5)].
	"if you know the index of each fields, use the faster atField: anInteger put: fieldValue, instead of atFieldName: anInteger put: fieldValue"
	dbf saveRecord. "make sure the last record is also saved"
	dbf inspect. "and dbf contents"
	dbf close "dont forget, or do XBaseFile allInstances do: [:each | each close]"
	
======================================================================================================
 Instance Variables:
	filterArray		-- Array with Blocks|nil	-- blocks that return true for [] value: self will lead to skipping the record
	lastUpdate		-- Data					-- date of last save
	fileName		-- String				-- name of currently open file
	fileStream		-- StandardFileStream	-- the open stream
	headerSize		-- Integer				-- size of header in bytes
	nRecords		-- Integer				-- number of records
	currentRecordIndex	-- Integer			-- index of selected record
	deleted			-- Boolean				-- indicates whether selected record is in fact deleted
	modified		-- Boolean				-- indicates whether selected record is modified
	eof				-- Boolean				-- end of file condition
	bof				-- Boolean				-- begin of file condition
	recordBytes		-- Array with ByteArrays	-- bytes in the current record (each field a ByteArray)
	readBlocks		-- Ord.Coll. with Blocks	-- each BlockContext defines how to read a field
	writeBlocks		-- Ord.Coll. with Blocks	-- each BlockContext defines how to write a field
	fields			-- Array with Strings	-- each string defines the name of a field (in UPPERCASE)
	nFields			-- Integer				-- the number of fields
	fieldsFormats	-- Ord.Coll. with Arrays	-- each array with 4 elements: fieldtype, fieldsize, decimals, fieldoffset in record
	fieldIndices		-- Dictionary			-- linking name of field (keys), to index in fieldFormats (value)

======================================================================================================
Record Header Structure of a DBF file (see http://www.e-bachmann.dk/docs/xbase.htm):
byte
---------------------------------------------------------------------------
0	Version number
----------------------------------------------------------------------------
1	Date of last update - year (0=1900)
2	Date of last update - month
3	Data of last update - day of month
---------------------------------------------------------------------------
4	Number of records (32 bits)
5
6
7
---------------------------------------------------------------------------
8	Length of header structure (16 bits)
9	
---------------------------------------------------------------------------
10	Length of each record (16 bits)
11
---------------------------------------------------------------------------
12	- 31 Reserved, etc, etc
---------------------------------------------------------------------------
32	Field Descriptor array (max 128 fields)
	Per field 32 bytes (0 - 31), see description below
n	n = 32 + fields*32 (or in numbers of bytes: 33 + fields*32)
-----------------------------------------------------------------------------
n+1	Terminator (16r0D = 13)
-----------------------------------------------------------------------------
m	Database Container
m+263	(Something related to Visual FoxPro) WITHIN THE HEADER STRUCTURE!!
-----------------------------------------------------------------------------


Now the Field Descriptor array:
byte
0	Field name in ASCII
10	(terminated by 16r00)
-----------------------------------------------------------------------------
11	Field type (ASCII)
-----------------------------------------------------------------------------
12	Field data address
15	(IRRELEVANT?)
-----------------------------------------------------------------------------
16	Field length
-----------------------------------------------------------------------------
17	Decimal count
-----------------------------------------------------------------------------
18	(reserved for multi-user dBASE)
19
-----------------------------------------------------------------------------
20	Work area ID (16r01 in all dBASE III files)
-----------------------------------------------------------------------------
20	(reserved for multi-user dBASE)
21
-----------------------------------------------------------------------------
23	Flag for SET FIELDS
-----------------------------------------------------------------------------
24	RESERVED
30
-----------------------------------------------------------------------------
31	Index field flag (dBASE IV, 16r00 -> ignored)
-----------------------------------------------------------------------------

And finally, record structure:

0	Field deleted flag 16r2A=42 (deleted) or 16r20=32 (valid)
-----------------------------------------------------------------------------
	Data (ASCII)
-----------------------------------------------------------------------------
	Field deleted flag etc (second field)!

!XBaseFile class methodsFor: 'examples' stamp: 'jmb 10/26/2004 22:38'!
example1
	"some examples (replace by your pathnames)"
	"XBaseFile example1"
	"dont forget to send <close> to the inspected instance. Otherwise, execute cleanup code: ' XBaseFile allInstances do: [:each | each close]' "

"	^ (self on: 'c:\etcovers\smalland\pat.dbf') inspect"
"	^ (self on: 'c:\etcovers\b_hoe\pat.dbf') inspect"
	^ (self on: 'c:\etcovers\vole3\pat.dbf') inspect
"	^ (self on: 'C:\etarcview\bushcricket\withET_CLASS\smallstepsCONV.dbf') inspect"
"	^ (self on: 'C:\etarcview\bushcricket\withET_CLASS\test4.dbf') inspect"
"	^ (self on: 'C:\hansdocs\projecten\laura\hansaantal.dbf') inspect"! !

!XBaseFile class methodsFor: 'examples' stamp: 'jmb 11/26/2004 12:37'!
exampleContents
	"some examples (replace by your pathnames)"
	"self exampleContents"
	
"	(self recordsFromFile: 'c:\etcovers\smalland\pat.dbf') inspect"
"	(self recordsFromFile: 'c:\etcovers\b_hoe\pat.dbf') inspect"
"	(self recordsFromFile: 'C:\etarcview\bushcricket\withET_CLASS\smallstepsCONV.dbf') inspect"
	
"	(self recordsAndFieldsFromFile: 'c:\etcovers\smalland\pat.dbf') inspect"
"	(self recordsAndFieldsFromFile: 'c:\etcovers\b_hoe\pat.dbf') inspect"
"	(self recordsAndFieldsFromFile: 'C:\etarcview\bushcricket\withET_CLASS\smallstepsCONV.dbf') inspect"

"	^self recordsAndFieldsFromUser inspect"
	^self recordsFromUser inspect! !

!XBaseFile class methodsFor: 'instance creation' stamp: 'jmb 10/28/2004 17:54'!
fromUser
	"XBaseFile fromUser"
	"XBaseFile fromUser contents"	
	" XBaseFile allInstances do: [:each | each close] "

| menu result |
	menu := StandardFileMenu oldFileMenu: FileDirectory default withPattern: '*.dbf'.
	result := menu startUpWithCaption: 'Select DBF to open...'.
	result
		ifNil: [^ nil].
	^self on: ( result directory fullNameFor: result name)! !

!XBaseFile class methodsFor: 'instance creation' stamp: 'jmb 10/26/2004 22:37'!
on: fileName 
	"open a DBFile object on an existing .DBF file, named fileName (String or Symbol, optionally including file path"

	^(super basicNew setName: fileName asString) open! !

!XBaseFile class methodsFor: 'instance creation' stamp: 'jmb 11/3/2004 22:11'!
on: fileName fields: fields
	"create a new DBFile object, with an OPEN file stream"
	"fileName sets the name of the file. Optionally it includes the file path. Fields is a sequential collection with Arrays defining name and format of each field:
		1) name, aString (automatically converted into UPPERCASE)
		2) a character indicating field type ($C -> Character, $N Numeric, $L Logical, $D Date)
		3) field length
		4) decimal positions (not relevant if field is not Numeric)
		5) nil (reserved for some? use)
	e.g. Array with: 'INTERNAL_ID' with: $N with: 5 with: 0 with: nil"

	^(super basicNew setName: fileName asString) create: fields! !

!XBaseFile class methodsFor: 'file contents' stamp: 'jmb 11/26/2004 12:32'!
recordsAndFieldsFromFile: aString 
	"retrieve all records from file named aString. As an additional record, 
	add the field labels. Close the file before returning"
	| file records |
	file := XBaseFile on: aString.
	[Cursor read
		showWhile: [records := file contentsWithFields]]
		ensure: [file close].
	^ records! !

!XBaseFile class methodsFor: 'file contents' stamp: 'jmb 11/26/2004 12:23'!
recordsAndFieldsFromUser
	"answer the records in a user-selected file. As an additional record, add the field labels. File will be closed before returning"
	"XBaseFile recordsAndFieldsFromUser"
	| menu result |
	menu := StandardFileMenu oldFileMenu: FileDirectory default withPattern: '*.dbf'.
	result := menu startUpWithCaption: 'Select DBF to open...'.
	result
		ifNil: [^ #()].
	^ self
		recordsAndFieldsFromFile: (result directory fullNameFor: result name)! !

!XBaseFile class methodsFor: 'file contents' stamp: 'jmb 11/26/2004 12:32'!
recordsFromFile: aString 
	"retrieve all records from file named aString. Close the file before 
	returning "
	| file records |
	file := XBaseFile on: aString.
	[Cursor read
		showWhile: [records := file contents]]
		ensure: [file close].
	^ records! !

!XBaseFile class methodsFor: 'file contents' stamp: 'jmb 11/26/2004 12:10'!
recordsFromUser
	"answer the records in a user-selected file. File will be closed before returning"
	"XBaseFile recordsFromUser"
	| menu result |
	menu := StandardFileMenu oldFileMenu: FileDirectory default withPattern: '*.dbf'.
	result := menu startUpWithCaption: 'Select DBF to open...'.
	result
		ifNil: [^ #()].
	^ self
		recordsFromFile: (result directory fullNameFor: result name)! !

!XBaseFile class methodsFor: 'testing' stamp: 'jmb 11/4/2004 17:39'!
test
	"create a dbf; add some filters"
	"XBaseFile test"
	"cleanUp code: 'XBaseFile allInstances do: [:each | each close]'"

	| a b |
	b := OrderedCollection new.
	Transcript cr; show: 'Building new test xBase file ...'.
	b add: #('abc' $N 5 0 nil).
	a := XBaseFile on: #test fields: b. 
	1 to: 10
		do: 
			[:ix | 
			a append;
				atFieldName: 'ABC' put: ix "printString"].
	Transcript show: ' Done'.
	Transcript cr; show: 'Records in file ' , a records printString.
	Transcript cr; show: 'Setting filter on field ABC. Filter expression ''((dbf atFieldName: #abc) asNumber  == 2) not '' ...'.
	a setFilter: [:dbf | ((dbf atFieldName: #abc) asNumber  == 2) not ].
	Transcript show: ' Done'.
	Transcript cr; show: 'Start 1st test: Skipping forward from top ... '.
	a do: [:dbf | Transcript cr; show: dbf currentRecordIndex printString , ' -- ' , (dbf atFieldName: 'ABC') printString].
	Transcript show: ' Done'.
	Transcript cr; show: 'Start 2st test: Skipping backward from bottom ... '.
	a reverseDo: [:dbf | Transcript cr; show: dbf currentRecordIndex printString , ' -- ' , (dbf atFieldName: #abc) printString].
	Transcript cr; show: ' Done'.
	Transcript cr; show: 'Closing xBase test file and exiting ...'.
	a close.
	Transcript cr; show: ' Done'.
	^a! !

!XBaseFile class methodsFor: 'testing' stamp: 'jmb 11/4/2004 17:44'!
testAfterEditing
	"open the file that was created programmatically with #testEditing"

	"XBaseFile testAfterEditing"

	"cleanUp code: 'XBaseFile allInstances do: [:each | each close]'"

	| dbf |
	dbf := XBaseFile on: 'testedit.dbf'.
	dbf halt.
	dbf close! !

!XBaseFile class methodsFor: 'testing' stamp: 'jmb 11/4/2004 17:42'!
testEditing
	"create a dbf with fieldtypes Number, Character, Float and Logical; add 10 records"
	"XBaseFile testEditing"
	"cleanUp code: 'XBaseFile allInstances do: [:each | each close]'"

	| a b rng |
	rng := Random new.
	rng seed: 2345678901.
	b := OrderedCollection new.
	Transcript cr; show: 'Building new test xBase file ...'.
	b addLast: #( 'abc' $N 5 0 nil).
	b addLast: #( 'et_class' $C 12 0 nil).
	b addLast: #( 'perimeter' $F 20 3 nil).
	b addLast: #( 'logical' $L 1 0 nil).
	a := XBaseFile on: 'testEdit' fields: b. 
	1 to: 10
		do: 
			[:ix | 
			a append.
			a atFieldName: 'ABC' put: ix.
			a atFieldName: 'ET_CLASS' put: 'MyClass', ix printString.
			a atFieldName: 'PERIMETER' put: ix  * 123456.12345.
			a atFieldName: 'LOGICAL' put: (rng next > 0.5)].
	Transcript show: ' Done'.
	Transcript cr; show:  a records printString , 'Records in file '.
	Transcript cr; show: 'Start 1st test: Skipping forward from top ... '.
	a do: [:dbf | Transcript cr;
				show: dbf currentRecordIndex printString;
				tab; show:  (dbf atFieldName: 'ABC') printString;
				tab; show:  (dbf atFieldName: 'ET_CLASS') printString;
				tab; show:  (dbf atFieldName: 'PERIMETER') printString;
				tab; show:  (dbf atFieldName: 'LOGICAL') printString].
	Transcript cr; show: ' Done'. 
	Transcript cr; show: 'Closing xBase test file and exiting ...'.
	a close.
	Transcript show: ' Done'.
	^a! !

!XBaseFile class methodsFor: 'testing' stamp: 'jmb 1/5/2005 23:00'!
testFloats
	"create a dbf with Float values"
	"XBaseFile testFloats"
	"cleanUp code: 'XBaseFile allInstances do: [:each | each close]'"
	"NOTE: floats that after formatting do not fit into the field can not be trusted!!
	It is the printString of the float that is stored, and it may use scientific notation" 

	| a b |
	b := OrderedCollection new.
	Transcript cr; show: 'Building new test xBase file ...'.
	b addLast: #( 'short' $F 18 0 nil). 
	b addLast: #( 'medium' $F 15 3 nil).
	b addLast: #( 'long' $F 20 3 nil).
	b addLast: #( 'num' $N 20 0 nil).
	b addLast: #( 'numdec' $N 20 3 nil).	
	a := XBaseFile on: 'testFloats' fields: b. 
	#(1234567.00008912345 1234567.00089123456 1234567.00891234567 1234567.08912345678 1234567.89123456789) 
		do: 
			[:ix | 
			a append.
			a atFieldName: 'SHORT' put: ix.
			a atFieldName: 'MEDIUM' put: ix.
			a atFieldName: 'LONG' put: ix.
			a atFieldName: 'NUM' put: ix.
			a atFieldName: 'NUMDEC' put: ix.						
			].
	Transcript show: ' Done'.
	Transcript cr; show:  a records printString , 'Records in file '.
	a do: [:dbf | Transcript cr;
				show: dbf currentRecordIndex printString;
				tab; show:  (dbf atFieldName: 'SHORT') printString;
				tab; show:  (dbf atFieldName: 'MEDIUM') printString;
				tab; show:  (dbf atFieldName: 'LONG') printString;
				tab; show:  (dbf atFieldName: 'NUM') printString;
				tab; show:  (dbf atFieldName: 'NUMDEC') printString							
				].
	Transcript cr; show: ' Done'. 
	Transcript cr; show: 'Closing xBase test file and exiting ...'.
	a close.
	Transcript show: ' Done'.
	^a! !

!XBaseFile class methodsFor: 'testing' stamp: 'jmb 1/6/2005 16:03'!
testIntegers
	"XBaseFile testIntegers"
	"cleanUp code: 'XBaseFile allInstances do: [:each | each close]'"
	"TAKE CARE TO MAKE THE FIELDSIZE LARGE ENOUGH!!!!!!!!"

	| a b |
	b := OrderedCollection new.
	Transcript cr; show: 'Building new test xBase file ...'.
	b addLast: #( 'short' $F 32 0 nil).
	b addLast: #( 'medium' $F 32 3 nil).
	b addLast: #( 'long' $F 32 3 nil).
	b addLast: #( 'num' $N 32 0 nil).
	b addLast: #( 'numdec' $F 32 3 nil).
	a := XBaseFile on: 'testIntegers' fields: b. 
	#(12345678901234 12345678901234567 123456789012345678 123456789012345678912 12345678901234567891234 123456789012345678912345678 )
		do: 
			[:ix | 
			a append.
			a atFieldName: 'SHORT' put: ix.
			a atFieldName: 'MEDIUM' put: ix.
			a atFieldName: 'LONG' put: ix.
			a atFieldName: 'NUM' put: ix.
			a atFieldName: 'NUMDEC' put: ix.
].
	Transcript show: ' Done'.
	Transcript cr; show: 'Records in file ' , a records printString.
	a do: [:dbf | Transcript cr;
				show: dbf currentRecordIndex printString;
				tab; show:  (dbf atFieldName: 'SHORT') printString;
				tab; show:  (dbf atFieldName: 'MEDIUM') printString;
				tab; show:  (dbf atFieldName: 'LONG') printString;
				tab; show:  (dbf atFieldName: 'NUM') printString;
				tab; show:  (dbf atFieldName: 'NUMDEC') printString].
	Transcript cr; show: 'Closing xBase test file and exiting ...'.
	a close.
	Transcript show: ' Done'.
	^a! !

!XBaseFile methodsFor: 'filters' stamp: 'jmb 10/26/2004 22:08'!
addFilter: aBlock 
	"Append a new filter condition. Filter conditions are chained by a logical AND"

	filterArray isNil ifTrue: [^self setFilter: aBlock].
	filterArray addLast: aBlock! !

!XBaseFile methodsFor: 'accessing - editing' stamp: 'jmb 11/3/2004 22:52'!
append
	"append a new, blank record. Committed directly to fileStream"
	"Add the end-of-file tag to the fileStream, and update header info in the fileStream"

	self modificationTest.
	"go to the end of the file (to the eof flag I assume)"
	nRecords > 0 
		ifTrue: 
			[self gotoRecord: nRecords.
			fileStream skip: recordSize]
		ifFalse: [fileStream position: headerSize].
	"increment nRecords and currentRecordIndex"
	nRecords := nRecords + 1.
	self currentRecordIndex: nRecords.
	"make new record and commit it to the file stream"
	self blankRecord; saveRecord.
	self appendEndOfFileFlag.
	self updateHeaderInfo	"currently the only possible change in headerinfo is in <records>"! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:15'!
appendEndOfFileFlag
	"set the end of file flag after the last record"

	fileStream position: headerSize + ( nRecords * recordSize).
	^fileStream nextPut: 26! !

!XBaseFile methodsFor: 'accessing' stamp: 'jmb 10/26/2004 21:54'!
atField: anInteger 
	"answer content of field at index anInteger in the current record"
	"assumes that <recordBytes> is already filled with the bytes of the current record"

	self deleted ifTrue: [^nil].
	^(readBlocks at: anInteger) value: self value: (recordBytes at: anInteger)! !

!XBaseFile methodsFor: 'accessing - editing' stamp: 'jmb 10/29/2004 13:14'!
atField: anInteger put: fieldValue 
	" set value of field at index anInteger to fieldValue"
	"Not committed on fileStream until #saveRecord is sent"

	| old bytes |
	old := recordBytes at: anInteger.
	bytes := (writeBlocks at: anInteger) value: self value: fieldValue.
	old = bytes 	"#(10 20 30 40 50) = #(10 20 30 40 50) true "
		ifFalse: 
			[self modified: true.
			recordBytes at: anInteger put: bytes]! !

!XBaseFile methodsFor: 'accessing - by name' stamp: 'jmb 10/29/2004 14:06'!
atFieldName: aFieldName 
	"answer the value of aFieldName in the current record <recordBytes>"

	| k |
	k := aFieldName asString asUppercase.
	^self atField: (fieldIndices at: k ifAbsent: [^nil])! !

!XBaseFile methodsFor: 'accessing - by name' stamp: 'jmb 10/29/2004 14:06'!
atFieldName: fieldName put: fieldValue 
	"set the value of aFieldName in the current record <recordBytes>"

	| k |
	k := fieldName asString asUppercase.
	^self atField: (fieldIndices at: k ifAbsent: [^nil]) put: fieldValue! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:15'!
blankRecord
	"current record is a new record with blank contents and the delete-flag set to false"

	"note that contents are not yet committed to the fileStream (requires a #saveRecord)"

	recordBytes := Array new: nFields.
	1 to: nFields
		do: [:i | recordBytes at: i put: ((writeBlocks at: i) value: self value: nil)].
	self deleted: false! !

!XBaseFile methodsFor: 'testing' stamp: 'jmb 10/26/2004 22:13'!
bof
	"test a beginning of file condition"

	^bof! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/29/2004 13:23'!
bytes2Num: aByteArray 
	"converts hex num in aByteArray into an integer"

	| aNum ix |
	aNum := 0.
	ix := 1.
	aByteArray do: 
			[:a | 
			aNum := aNum + (a * ix).
			ix := ix * 256].
	^aNum! !

!XBaseFile methodsFor: 'open/close' stamp: 'jmb 11/3/2004 22:52'!
close
	" Close XBaseFile flushing last updates and release object"

	self modificationTest.			"current record is saved"
	fileStream isNil ifFalse: [fileStream close].
	fileStream := nil! !

!XBaseFile methodsFor: 'accessing' stamp: 'jmb 11/3/2004 22:53'!
contents
	"dump all data in anArray excluding header!!"
	"deleted records are included as empty Arrays (all nil)"

	| contents |
	self modificationTest.
	contents := Array new: self records.
	1 to: nRecords
		do: 
			[:i | 
			fileStream position: headerSize + ((i - 1) * recordSize).
			self currentRecordIndex: i.
			self getRecord.
			contents at: i put: self recordData].
	^contents! !

!XBaseFile methodsFor: 'accessing' stamp: 'jmb 11/3/2004 22:51'!
contentsAtField: aKey 
	"dump data (all records, including deleted) found at field aKey in anArray (=column)"

	"aKey itself is not included in anArray"

	| contents index |
	index := fieldIndices at: aKey ifAbsent: [^#()].
	self modificationTest.
	contents := Array new: self records.
	1 to: nRecords
		do: 
			[:i | 
			fileStream position: headerSize + ((i - 1) * recordSize).
			self currentRecordIndex: i.
			self getRecord.
			contents at: i put: (self atField: index)].
	^contents! !

!XBaseFile methodsFor: 'accessing' stamp: 'jmb 11/26/2004 12:22'!
contentsWithFields
	"dump all data in anArray including the field labels (as a last record)"
	"deleted records are included as empty Arrays (all nil)"

	| contents |
	self modificationTest.
	contents := Array new: self records + 1.
	1 to: nRecords
		do: 
			[:i | 
			fileStream position: headerSize + ((i - 1) * recordSize).
			self currentRecordIndex: i.
			self getRecord.
			contents at: i put: self recordData].
	contents at: self records + 1 put: self fields.
	^contents! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/29/2004 13:25'!
create: fDictionary 
	"create a new XBaseFile (with open file stream), with fields defined by fDictionary, and 0 records"
	| pos |
	fileStream := (StandardFileStream newFileNamed: fileName) binary.
	lastUpdate := Date today.
	nRecords := 0.
	currentRecordIndex := 0.
	self modified: false.
	self deleted: false.
	pos := 1.
	fields := OrderedCollection new.
	fieldIndices := Dictionary new.
	fieldFormats := OrderedCollection new.
	readBlocks := OrderedCollection new.
	writeBlocks := OrderedCollection new.
	fDictionary
		do: [:a | 
			| name format | 
			name := a first asString asUppercase.
			name := name
						copyFrom: 1
						to: (10 min: name size).
			"10 bytes for the name (11th byte should be 16r00=0)"
			fields addLast: name.
			format := a copyFrom: 2 to: a size.
			format at: 4 put: pos.
			"fill in the missing offset value (at pos=0 we have the deleted 
			flag) "
			fieldIndices at: name put: fields size.
			fieldFormats addLast: format.
			readBlocks
				addLast: (self
						readBlockFor: (format at: 1)
						dec: (format at: 3)).
			writeBlocks
				addLast: (self
						writeBlockFor: (format at: 1)
						dec: (format at: 3)
						length: (format at: 2)).
			pos := pos
						+ (format at: 2)].
	fields := fields asArray.
	nFields := fields size.
	recordSize := pos.
	headerSize := 34 + (fields size * 32).
	self saveHeaderInfo! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:19'!
currentRecordIndex
	"answer the index of the current record"

	^currentRecordIndex! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:19'!
currentRecordIndex: anInteger
	"keep the index of the current record in <currentRecordIndex>"

	currentRecordIndex := anInteger! !

!XBaseFile methodsFor: 'accessing - editing' stamp: 'jmb 10/26/2004 22:02'!
delete
	"mark as deleted current record. Not committed until #saveRecord is sent"

	self deleted 
		ifFalse: 
			[self deleted: true.
			self modified: true]! !

!XBaseFile methodsFor: 'testing' stamp: 'jmb 10/26/2004 22:13'!
deleted
	"true if the current record has the deleted-tag"
	"deleted is set with #getRecord (read from fileStream). If the record is deleted (#delete) or restored (#restore), it is changed using #deleted: aBoolean. 
	When committed to fileStream, the value of <deleted> is used as the deleted tag"

	^deleted! !

!XBaseFile methodsFor: 'testing' stamp: 'jmb 10/26/2004 22:13'!
deleted: aBoolean
	"set the deleted-tag for the current record (not yet committed to the fileStream though!!)"

	deleted := aBoolean! !

!XBaseFile methodsFor: 'accessing - asynchronous' stamp: 'jmb 10/26/2004 21:58'!
deletedFlagRecord: anInteger 
	"answer the value of the deleted flag for record anInteger"
	"16r2A (42) deleted; 16r20 (32) valid"

	fileStream position: headerSize + ((anInteger - 1) * recordSize).
	^fileStream next! !

!XBaseFile methodsFor: 'iterating' stamp: 'jmb 10/26/2004 22:09'!
do: aBlock 
	"iterate over all records. Evaluate aBlock with self as the argument, each time currentRecord points to a new record"

	self goTop.
	[self eof or: [self bof]] whileFalse: 
			[aBlock value: self.
			self skip]! !

!XBaseFile methodsFor: 'accessing - asynchronous' stamp: 'jmb 10/26/2004 21:58'!
endOfFileFlag
	"answer the last byte of the file, the end of file flag 16r1A (26)"

	fileStream position: headerSize + ( nRecords * recordSize).
	^fileStream next! !

!XBaseFile methodsFor: 'testing' stamp: 'jmb 10/26/2004 22:14'!
eof
	"test an end of file condition"

	^eof! !

!XBaseFile methodsFor: 'filters' stamp: 'jmb 10/26/2004 22:08'!
evaluateFilters
	"answer true if the evaluated value should be trapped by the filter, and the record skipped"

	| answ |
	answ := true.
	filterArray do: [:a | answ := answ and: [a value: self]].	"logically AND evaluations of all elements"
	^answ not! !

!XBaseFile methodsFor: 'accessing - field format' stamp: 'jmb 10/29/2004 14:02'!
fieldDec: fieldName 
	"returns field decimal positions for field named fieldName (string or 
	symbol, case-insensitive)"
	^ (fieldFormats
		at: (fieldIndices
				at: fieldName asString asUppercase
				ifAbsent: [^ nil]))
		at: 3! !

!XBaseFile methodsFor: 'accessing - field format' stamp: 'jmb 10/29/2004 14:02'!
fieldInfos: fieldName 
	"returns an array defining the format of field named fieldName (string or symbol, case insensitive): 
	1 = field type; 2 = field length; 3 = field decimal positions; 4 = field offset inside record.  
	Method returns nil if fieldName is not in file"
	^ fieldFormats
		at: (fieldIndices
				at: fieldName asString asUppercase
				ifAbsent: [^ nil])! !

!XBaseFile methodsFor: 'accessing - field format' stamp: 'jmb 10/29/2004 13:57'!
fieldInfosDictionary
	"returns a Dictionary with the field info. Key is the field name, value is an array with field type, field length, field decimal positions and field offset inside record"

	| dict |
	dict := Dictionary new.
	1 to: fields size
		do: [:i | dict at: (fields at: i) put: (fieldFormats at: i)].
	^dict! !

!XBaseFile methodsFor: 'accessing - field format' stamp: 'jmb 10/29/2004 14:03'!
fieldLen: fieldName 
	"returns field length for field named fieldName (string or symbol, case 
	insensitive). Method returns nil if fieldName is not in file"
	^ (fieldFormats
		at: (fieldIndices
				at: fieldName asString asUppercase
				ifAbsent: [^ nil]))
		at: 2! !

!XBaseFile methodsFor: 'accessing - field format' stamp: 'jmb 10/29/2004 14:03'!
fieldOffset: fieldName 
	"returns the offset of the field in the record. Method returns nil if fieldName is not in file"
	^ (fieldFormats
		at: (fieldIndices
				at: fieldName asString asUppercase
				ifAbsent: [^ nil]))
		at: 4! !

!XBaseFile methodsFor: 'accessing - field format' stamp: 'jmb 10/26/2004 22:04'!
fieldType: fieldName 
	"returns aCharacter describing field type of field named fieldName (string or symbol, case insensitive). $C = Character field, $N = Numeric field, $L = logical field, $D = date field, $F = Float. Returns nil if fieldName is not in file"

	^(fieldFormats at: (fieldIndices at: fieldName asString asUppercase ifAbsent: [^nil])) at: 1! !

!XBaseFile methodsFor: 'accessing' stamp: 'jmb 10/26/2004 21:56'!
fields
	"answer anArray with field labels (all uppercase) "
	^ fields! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/29/2004 13:27'!
formatBooleanFieldValue: v 
	"convert aBoolean into the appropriate String"
	" 'F' asString = 'F' true "
	" 'F' asByteArray #[70]; 'T' asByteArray #[84] "

	| answ |
	v isNil ifTrue: [^'F'].
	answ := v ifTrue: ['T'] ifFalse: ['F'].
	^answ! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/29/2004 13:28'!
formatDateFieldValue: v 
	"convert aDate into the appropriate String"
	| answ tmp |
	v isNil 
		ifTrue: 
			[^self formatDateFieldValue: (Date 
						year: 1
						month: 1
						day: 1)].
	answ := v year printString.
	tmp := '0' , v monthIndex printString.
	tmp := tmp copyFrom: tmp size - 1 to: tmp size.
	answ := answ , tmp.
	tmp := '0' , v day printString.
	tmp := tmp copyFrom: tmp size - 1 to: tmp size.
	answ := answ , tmp.
	^answ! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 1/6/2005 16:01'!
formatNumberFieldLength: l decimalC: d value: v 
	"convert a Number into the appropriate ByteString"
	| space stream value result |
	space := Character space.
	v isNil ifTrue: [^String new: l withAll: space].
	stream := WriteStream on: (String new: l).
	d > 0 
		ifFalse: 
			[value := v asInteger.
			value printOn: stream]
		ifTrue: 
			[value := v asFloat roundTo: (1 / (10 raisedToInteger: d)) asFloat.
			value printOn: stream].
	result := stream contents.
	result size < l 
		ifTrue: [l - result size timesRepeat: [stream nextPut: space]]
		ifFalse: [result size > l ifTrue: [self halt]].
	^stream contents! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 1/6/2005 16:00'!
formatNumberFieldLengthOld: l decimalC: d value: v 
	"convert a Number into the appropriate ByteString"

	| int is sgn comma cti tmp dcs answ space |
	space := Character space.
	v isNil ifTrue: [^String new: l withAll: space].
	int := v asInteger.
	is := int abs printString.
	sgn := int < 0 ifTrue: ['-'] ifFalse: [''].
	comma := d > 0 ifTrue: ['.'] ifFalse: [''].
	cti := (tmp := l - d - is size - sgn size - comma size) > 0 
				ifTrue: [String new: tmp withAll: space]
				ifFalse: ['']. "self halt."
	d > 0 
		ifTrue: 
			[| value stream result |
			value := (v roundTo: 1 / (10 raisedToInteger: d)) asFloat.
			stream := WriteStream on: (String new: l).
			value printOn: stream.
			result := stream contents.
			result size < l 
				ifTrue: [l - result size timesRepeat: [stream nextPut: space]. ^ stream contents]
				ifFalse: [result size > l ifTrue: [self error: 'value does not fit into field' "result := result allButFirst: result size - l" "arbitrary choose to remove chars at front"]].
			^result]
		ifFalse: [dcs := ''].
	answ := sgn , cti , is , comma , dcs.
	answ := answ copyFrom: answ size - l + 1 to: answ size.
	^answ! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/29/2004 13:31'!
formatStringFieldLength: len value: v 
	"convert a String into the appropriate ByteString"
	| answ space |
	space := Character space.
	v isNil ifTrue: [^ String new: len withAll: space].
	answ := v , (String new: len withAll: space) copyFrom: 1 to: len.
	^ answ! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:21'!
getDBFInfos
	"get all infos about open DBF file"

	self
		getHeaderInfo;
		getFieldsInfos! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/28/2004 17:48'!
getFieldsInfos
	"extract field formats from the file stream and move these into a fieldsInfos dictionary"
	"keep also a fields array, with just the names in the correct order"

	| pos name type len dec fpos |
	fields := OrderedCollection new.
	pos := 32.
	fieldIndices := Dictionary new.
	fieldFormats := OrderedCollection new.
	readBlocks := OrderedCollection new.
	writeBlocks := OrderedCollection new.
	fpos := 1.
	[pos < (headerSize - 2)]
		whileTrue: 
			[fileStream position: pos.
			name := (fileStream upTo: 0) "asByteString in VW" asString.
			fileStream position: pos + 11.
			type := fileStream next asCharacter.
			fileStream skip: 4.
			dec := 0.
			type == $C
				ifTrue: [len := self bytes2Num: (fileStream next: 2)]
				ifFalse: 
					[len := fileStream next.
					dec := fileStream next].
			fields addLast: name.
			fieldIndices at: name put: fields size.
			fieldFormats addLast: (Array with: type with: len with: dec with: fpos).
			readBlocks addLast: (self readBlockFor: type dec: dec).
			writeBlocks addLast: (self writeBlockFor: type dec: dec length: len).
			fpos := fpos + len.
			pos := pos + 32].
	fields := fields asArray.
	nFields := fields size! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/28/2004 16:55'!
getHeaderInfo
	"extract header information and move these data to instance variables"

	| d m y |
	fileStream position: 1.
	y := fileStream next.
	m := fileStream next.
	d := fileStream next.
	y < 1900 ifTrue: [y := 1900 + y].
"	lastUpdate := Date newDay: d monthNumber: m year: y."
	lastUpdate := Date year: y month: m day: d.
	nRecords := self bytes2Num: (fileStream next: 4).
	headerSize := self bytes2Num: (fileStream next: 2).
	recordSize := self bytes2Num: (fileStream next: 2)! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:22'!
getRecord
	"extract current record contents and move these to fieldsContents values"

	"requires that fileStream is already positioned at the beginning of the record"

	recordBytes := Array new: nFields.
	self deleted: fileStream next = 42.	"first Byte is deleted-flag. 16r20 32 for valid, and 16r2A 42 for deleted"
	1 to: nFields
		do: [:i | recordBytes at: i put: (fileStream next: ((fieldFormats at: i) at: 2))]! !

!XBaseFile methodsFor: 'positioning' stamp: 'jmb 10/29/2004 13:36'!
goBottom
	"set last record in file as current record. Deleted and Filter settings are accounted for"

	self
		goto: nRecords;
		skip: 0 direction: -1! !

!XBaseFile methodsFor: 'positioning' stamp: 'jmb 10/29/2004 13:36'!
goTop
	"set first record in file as current record. Deleted and Filter settings are accounted for"

	self
		goto: 1;
		skip: 0 direction: 1! !

!XBaseFile methodsFor: 'positioning' stamp: 'jmb 11/3/2004 22:52'!
goto: rNumber 
	"position file on record number rNumber"

	"add a blank record when rNumber exceeds <records>. Do not commit anything to the fileStream yet (except for the previously modified record)"

	| recordNumber |
	eof := false.
	bof := false.
	self modificationTest.
	rNumber < 1 
		ifTrue: 
			[recordNumber := 1.
			bof := true]
		ifFalse: [recordNumber := rNumber].
	recordNumber <= nRecords 
		ifTrue: 
			[self gotoRecord: recordNumber.
			self getRecord]
		ifFalse: 
			[self currentRecordIndex: nRecords + 1. "increase record count by 1"
			eof := true.
			self blankRecord	"a blank record becomes the current record"]! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:22'!
gotoRecord: recordNumber 
	"position currentRecord to recordNumber"

	(recordNumber <= nRecords and: [recordNumber > 0]) 
		ifFalse: [^self error: 'Record number outside file bounds'].
	fileStream position: headerSize + ((recordNumber - 1) * recordSize).
	self currentRecordIndex: recordNumber! !

!XBaseFile methodsFor: 'accessing - asynchronous' stamp: 'jmb 10/26/2004 21:58'!
headerSize
	fileStream position: 8.
	^self bytes2Num: (fileStream next: 2)! !

!XBaseFile methodsFor: 'testing' stamp: 'jmb 10/26/2004 22:14'!
isDBFFile
	"tests if the open file is a .DBF file (signature byte=03)"

	fileStream position: 0.
	^fileStream next asInteger = 3! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 11/3/2004 22:51'!
modificationTest
	"if changed, save the current record"
	self modified ifTrue: [self saveRecord].! !

!XBaseFile methodsFor: 'testing' stamp: 'jmb 10/26/2004 22:14'!
modified
	"true if the current record has been changed"

	^modified! !

!XBaseFile methodsFor: 'testing' stamp: 'jmb 10/26/2004 22:14'!
modified: aBoolean
	"set the modified tag"

	^modified := aBoolean! !

!XBaseFile methodsFor: 'filters' stamp: 'jmb 10/26/2004 22:08'!
noFilter
	"remove any filter setting for file"

	filterArray := nil! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:22'!
num2Bytes: aNum size: aSize 
	"converts aNum hex num to aByteArray"

	| tmp answ |
	answ := ByteArray new: aSize withAll: 32.
	tmp := aNum.
	1 to: aSize
		do: 
			[:ix | 
			answ at: ix put: ("Character value:" tmp \\ 256).
			tmp := tmp // 256].
	^answ! !

!XBaseFile methodsFor: 'accessing - asynchronous' stamp: 'jmb 10/26/2004 21:59'!
numberOfRecords
	fileStream position: 4.
	^self bytes2Num: (fileStream next: 4)! !

!XBaseFile methodsFor: 'accessing - asynchronous' stamp: 'jmb 10/26/2004 21:59'!
offsetRecord: anInteger 
	"answer the offset (position) where record anInteger starts"

	^headerSize + ((anInteger - 1) * recordSize)! !

!XBaseFile methodsFor: 'open/close' stamp: 'jmb 10/28/2004 17:01'!
open
	"Open an existing .DBF file in binary mode"
	"This methods is called by class methods #new: and  
	#new:fields:. May be called directly to reopen a closed file"
	fileStream := (StandardFileStream fileNamed: fileName) binary.
	self isDBFFile
		ifTrue: [self modified: false.
			self getDBFInfos.
			self goto: 1.
			^ self]
		ifFalse: [fileStream close.
			^ self error: fileName , ': is not a .DBF file']! !

!XBaseFile methodsFor: 'debugging' stamp: 'jmb 10/26/2004 22:07'!
printTranscript
	"print all records on the Transcript"

	self goTop.
	Transcript cr.
	fields do: [:each | Transcript show: each] separatedBy: [Transcript tab].
	[eof or: [bof]] whileFalse: 
			[Transcript cr.
			fields do: [:each | Transcript show: (self atFieldName: each) printString]
				separatedBy: [Transcript tab].
			self skip]! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/26/2004 22:04'!
readBlockFor: aSymbol dec: dec

	aSymbol == $N ifTrue: [^self readBlockForNumber].
	aSymbol == $C ifTrue: [^self readBlockForString].
	aSymbol == $L ifTrue: [^self readBlockForBoolean].
	aSymbol == $D ifTrue: [^self readBlockForDate].
	aSymbol == $F ifTrue: [^self readBlockForFloatDec: dec].! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/26/2004 22:05'!
readBlockForBoolean

	^[:dbf :bytes | dbf unformatBooleanFromBytes: bytes]! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/28/2004 17:49'!
readBlockForDate

	^[:dbf :bytes | dbf unformatDateFromString: bytes "asByteString in VW" asString]! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/26/2004 22:51'!
readBlockForFloat
	^ [:dbf :bytes | dbf unformatFloatFromString: bytes "asByteString" asString]! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/26/2004 22:51'!
readBlockForFloatDec: anInteger 
	^ [:dbf :bytes | dbf unformatFloatFromString: bytes asString
	"asByteString" dec: anInteger]! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/26/2004 22:49'!
readBlockForNumber

	^[:dbf :bytes | dbf unformatNumberFromString: bytes "asByteString" asString]! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/26/2004 22:52'!
readBlockForString
	^ [:dbf :bytes | dbf unformatStringFromString: bytes asString
	"asByteString"]! !

!XBaseFile methodsFor: 'accessing' stamp: 'jmb 10/26/2004 21:57'!
recordData
	"answer anArray with all the data in the current record (as Smalltalk objects, not bytes)"

	| data |
	data := Array new: nFields.
	self deleted ifTrue: [^data].
	1 to: nFields
		do: 
			[:i | 
			data at: i put: ((readBlocks at: i) value: self value: (recordBytes at: i))].
	^data! !

!XBaseFile methodsFor: 'accessing - asynchronous' stamp: 'jmb 10/26/2004 21:59'!
recordLength
	"the length of each record, in bytes"

	fileStream position: 10.
	^self bytes2Num: (fileStream next: 2)! !

!XBaseFile methodsFor: 'accessing' stamp: 'jmb 10/26/2004 22:44'!
records
	"return the number of records in file (including records with deleted-tag)"

	^nRecords! !

!XBaseFile methodsFor: 'accessing - editing' stamp: 'jmb 10/26/2004 22:02'!
restore
	"unmark as deleted current record. Not committed until #saveRecord is sent"

	self deleted 
		ifTrue: 
			[self deleted: false.
			self modified: true]! !

!XBaseFile methodsFor: 'iterating' stamp: 'jmb 10/26/2004 22:09'!
reverseDo: aBlock 
	"iterate backwards over all records. Evaluate aBlock with self as the argument, each time currentRecord points to a new record"

	self goBottom.
	[self eof or: [self bof]] whileFalse: 
			[aBlock value: self.
			self skip: -1]! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:32'!
saveFieldsInfos
	"commit the field formats to the fileStream"
	| pos |
	pos := 32.
	fileStream position: 32.
	fieldFormats
		with: fields
		do: [:info :name | 
			fileStream position: pos.
			(ByteArray new: 11 withAll: 0)
				do: [:byte | fileStream nextPut: byte].
			"erase the previous contents"
			fileStream position: pos.
			name asByteArray
				do: [:byte | fileStream nextPut: byte].
			"Field name: 10 characters; 11th byte = 0"
			fileStream position: pos + 11.
			fileStream nextPut: (info at: 1) asInteger.
			"Field type"
			(info at: 1)
					== $C
				ifTrue: [fileStream position: pos + 16;
						
						nextPutAll: (self
								num2Bytes: (info at: 2)
								size: 2)]
				ifFalse: [fileStream position: pos + 16;
						
						nextPutAll: (self
								num2Bytes: (info at: 2)
								size: 1).
					"Field length"
					fileStream position: pos + 17;
						
						nextPutAll: (self
								num2Bytes: (info at: 3)
								size: 1)].
			fileStream position: pos + 20;
				
				nextPutAll: (self num2Bytes: 1 size: 1).
			"work area id - not really relevant"
			fileStream position: pos + 31;
				
				nextPutAll: (self num2Bytes: 0 size: 1).
			"ignore index field"
			pos := pos + 32].
	fileStream position: pos! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:32'!
saveHeaderInfo
	"save .DBF file header info's"

	fileStream
		position: 0;
		nextNumber: 1 put: 3;
		nextNumber: 1 put: lastUpdate year - 1900;
		nextNumber: 1 put: lastUpdate monthIndex;
		nextNumber: 1 put: lastUpdate dayOfMonth ;
		nextPutAll: (self num2Bytes: nRecords size: 4);
		nextPutAll: (self num2Bytes: headerSize size: 2);
		nextPutAll: (self num2Bytes: recordSize size: 2).
	self saveFieldsInfos.
	fileStream nextNumber: 1 put: 13.			"Terminator (of the header structure) 16r0D = 13"
	fileStream nextNumber: 1 put: 0.				"insert something, otherwise completely empty"
	fileStream nextNumber: 1 put: 26				"End of file flag 16r1A = 26"! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/29/2004 13:34'!
saveRecord
	"commit the current record contents to the fileStream (overwrite 
	previous contents)"
	"#saveRecord is automatically performed when moving to another record 
	(#goto:) and the current record has been modified. 
	see #append, #close and #goto:"
	self gotoRecord: currentRecordIndex.
	"position fileStream at offset record"
	self deleted
		ifTrue: [fileStream nextPut: 42]
		ifFalse: [fileStream nextPut: 32].
	"the deleted record flag"
	recordBytes
		do: [:fieldBytes | fieldBytes
				do: [:byte | fileStream nextPut: byte]].
	"pump record bytes into the stream"
	self modified: false! !

!XBaseFile methodsFor: 'filters' stamp: 'jmb 10/26/2004 22:09'!
setFilter: aBlock 
	"set a new filter condition. Any preceding filter condition is lost. aBlock is a one-argument Block expecting the receiver as its argument"

	filterArray := OrderedCollection with: aBlock! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:33'!
setName: aString 
	"Set the .DBF file name to aString"

	| ext |
	ext := ''.
	(aString includes: $.) ifFalse: [ext := '.DBF'].
	fileName := aString , ext! !

!XBaseFile methodsFor: 'accessing - asynchronous' stamp: 'jmb 10/26/2004 21:59'!
signature
	"answer the first byte"

	fileStream position: 0.
	^fileStream next! !

!XBaseFile methodsFor: 'positioning' stamp: 'jmb 10/29/2004 13:37'!
skip
	"set next record in file as current record. Deleted and Filter settings are accounted for"

	^self skip: 1 direction: 1! !

!XBaseFile methodsFor: 'positioning' stamp: 'jmb 10/29/2004 13:37'!
skip: aNumber 
	"jump to next aNumber record. Jump maybe forward (aNumber > 0) or backward (aNumber <0). Deleted and Filter settings are accounted for."

	self skip: aNumber direction: aNumber abs / aNumber! !

!XBaseFile methodsFor: 'positioning' stamp: 'jmb 10/26/2004 22:13'!
skip: aNumber direction: ofs 
	"skip aNumber of positions (records) in the direction ofs (+1 or -1)"
	"halt when bof or eof is encountered; deleted or filtered records do not count, I think"

	self goto: currentRecordIndex + aNumber.
	
	[(((filterArray isNil not and: [self evaluateFilters]) or: [deleted]) 
		and: [eof not]) and: [bof not]] 
			whileTrue: [self goto: currentRecordIndex + ofs]! !

!XBaseFile methodsFor: 'accessing - asynchronous' stamp: 'jmb 10/26/2004 21:59'!
testHeaderTerminator
	"check whether the Terminator is present"

	fileStream position: self headerSize - 1.
	^fileStream next = 16r0D "13"! !

!XBaseFile methodsFor: 'accessing - asynchronous' stamp: 'jmb 10/26/2004 22:00'!
testNumberOfAttributes
	"derive the number of attributes from the space used by the header (32 bytes per attribute)"

	^fields size = (self headerSize - 1 - 32 / 32) "1 = the terminator; 32 = the fields in the header before the attribute fields"! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:34'!
unformatBooleanFromBytes: aByteArray
	"Logic -> Boolean: $T asInteger = 84 and $F asInteger = 70"

	^aByteArray first = 84! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/28/2004 18:03'!
unformatDateFromString: aString 
	"Convert aString into aDate"

	| rStream y m dy |
	rStream := ReadStream 
				on: aString
				from: 1
				to: 8.
	y := (rStream next: 4) asInteger.
	m := (rStream next: 2) asInteger.
	dy := (rStream next: 2) asInteger.
	^Date 
		year: y
		month: m
		day: dy! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/29/2004 13:54'!
unformatFloatFromString: aString 
	"Convert aString into aNumber. For floating point values scan aString for possible scientific notation"
	"scientific notation, e.g. '6.665804E+001' "
	" '6.665804E+001' -> '6.665804e001' "

	"Integer readFrom: '23.456' readStream 23"
	"'23.456' asNumber 23.456"
	"'23.45678912345' asNumber 23.45678912345"
	"withBlanksTrimmed not required anymore in last line (spaces are filtered out)"

	| rStream wStream n sp |
	rStream := aString readStream.
	wStream := WriteStream on: (String new: aString size).
	sp := Character space.
	[rStream atEnd] whileFalse: 
			[n := rStream next.
			(n == $E or: [n == $e]) 
				ifTrue: [wStream nextPut: $e]
				ifFalse: [n == $+ ifFalse: [n == sp ifFalse: [wStream nextPut: n]]]].
	^wStream contents asNumber! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/28/2004 18:06'!
unformatFloatFromString: aString dec: dc
	"Convert aString into aFloating point; enforce dc as number of decimals"

	^(self unformatFloatFromString: aString) 
		roundTo: (1 / (10 raisedToInteger: dc) ) asFloat! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/28/2004 18:06'!
unformatNumberDecimals: dc fromString: aString 
	"Convert aString into aFloating point; enforce dc as number of decimals"

	^(self unformatFloatFromString: aString) 
		roundTo: (1 / (10 raisedToInteger: dc) ) asFloat! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 3/21/2005 12:56'!
unformatNumberFromString: aString 
	"Convert aString into aNumber. For floating point values scan aString 
	for possible scientific notation"
	"scientific notation, e.g. '6.665804E+001'"
	"'6.665804E+001' -> '6.665804d001'"
	"Integer readFrom: '23.456' readStream 23"
	"'23.456' asNumber 23.456"
	"'23.45678912345' asNumber 23.45678912345"
	"withBlanksTrimmed in last line not required anymore (at least, spaces 
	are ignored)"
	| rStream wStream n sp |
	aString first == $* ifTrue: [^nil]. "or byte = 42"
	rStream := aString readStream.
	wStream := WriteStream
				on: (String new: aString size).
	sp := Character space.
	[rStream atEnd]
		whileFalse: [n := rStream next.
			(n == $E
					or: [n == $e])
				ifTrue: [wStream nextPut: $e]
				ifFalse: [n == $+
						ifFalse: [n == sp
								ifFalse: [wStream nextPut: n]]]].
	^ wStream contents asNumber! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:36'!
unformatStringFromString: aString 
	"aString contains the data string; just remove trailing spaces"

	| space i |
	space := Character space.
	aString last == space ifFalse: [^aString].
	i := aString size.
	[i > 0 and: [(aString at: i) == space]] whileTrue: [i := i - 1].
	^aString copyFrom: 1 to: i! !

!XBaseFile methodsFor: 'private' stamp: 'jmb 10/26/2004 22:36'!
updateHeaderInfo
	"currently only the number of records may have changed (increased through #append)"

	fileStream
		position: 4;
		nextPutAll: (self num2Bytes: nRecords size: 4)! !

!XBaseFile methodsFor: 'testing' stamp: 'jmb 10/26/2004 22:15'!
version
	"current version of XBase"

	^ 'XBaseFile ReaderWriter by Hans Baveco (Hans.Baveco@wur.NL). A thorough rewrite of Antonio d''Avino (andavino@tin.it) Squeak XBaseFile'! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/26/2004 22:06'!
writeBlockFor: aSymbol dec: dec length: len

	aSymbol == $N ifTrue: [^self writeBlockForNumberDec: dec length: len].
	aSymbol == $C ifTrue: [^self writeBlockForStringLength: len].
	aSymbol == $L ifTrue: [^self writeBlockForBoolean].
	aSymbol == $D ifTrue: [^self writeBlockForDate].
	aSymbol == $F ifTrue: [^self  writeBlockForNumberDec: dec length: len]! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/29/2004 10:00'!
writeBlockForBoolean
"	^ [:dbf :val | (dbf formatBooleanFieldValue: val) asByteString] for VW"
		
	^ [:dbf :val | ByteArray withAll: (dbf formatBooleanFieldValue: val)] ! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/29/2004 10:01'!
writeBlockForDate
"	^ [:dbf :val | (dbf formatDateFieldValue: val) asByteString] for VW"
	^ [:dbf :val | ByteArray withAll: (dbf formatDateFieldValue: val)]! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/29/2004 10:02'!
writeBlockForNumberDec: d length: len 
"	^ [:dbf :val | (dbf formatNumberFieldLength: len decimalC: d value: val) asByteString] for VW"
	^ [:dbf :val | ByteArray withAll: (dbf formatNumberFieldLength: len decimalC: d value: val)]! !

!XBaseFile methodsFor: 'block contexts' stamp: 'jmb 10/29/2004 10:02'!
writeBlockForStringLength: len 
"	^ [:dbf :val | (dbf formatStringFieldLength: len value: val) asByteString] for VW"
	^ [:dbf :val | ByteArray withAll: (dbf formatStringFieldLength: len value: val)]! !
